// deHarry:
// Der Sketch basiert auf den Vorarbeiten von Slaven1337 im Arduino Forum:
//  https://forum.arduino.cc/t/smart-meter-sml-hex-code-zerlegen/526925/9
// Die nahezu vollständige Analyse des SML Telegramms von Stefan Weigert hat mir SEHR geholfen. Siehe:
//  http://www.stefan-weigert.de/php_loader/sml.php
//
// Slaven1337:
// Credits - This code is based on the work of user "rollercontainer" at 
//  https://www.photovoltaikforum.com/volkszaehler-org-f131/sml-protokoll-hilfe-gesucht-sml-gt-esp8266-gt-mqtt-t112216-s10.html
// For SML protocol see http://www.schatenseite.de/2016/05/30/smart-message-language-stromzahler-auslesen/
// Hardware used: https://www.msxfaq.de/sonst/bastelbude/smartmeter_d0_sml.htm 
// \Slaven1337
//
// deHarry, 2023-01-08 .. 2023-01-18
// Programmentwicklung und Debugging erfolgte auf einem Arduino Nano. Letzlich wird ein ESP-01 die ausgelesenen und dekodierten Daten
//	im WLAN zur Verfügung stellen. 
// Hardware:
// Adapter aus der Empfängerdiode einer Reflexlichtschranke in einem 3D gedruckten Halter mit zwei Magneten, Pluspol auf RxD (IO 0)
//  Achtung! Kein PullUp beim Wemos D1 R2, beim ESP-01 muss er 1 kΩ klein sein.
// Am ESP ist ein OLED Display per I2C angeflanscht, das per Knopfdruck aktiviert werden kann (normalerweise AUS um die Pixel zu schonen).
// Software:
// Einlesen und Parsing des Telegramms komplett umgebaut gegenüber Slaven1337. Shifterei durch "union" unnötig gemacht.
//  Achtung! Der Wert auf Index 11 der Vergleichsstrings hat bei meinem MT175 (Zweirichtungszähler) unterschiedliche Werte: 52, 82, A2. 
//	Scheinbar abhängig von der Tageszeit oder von Bezug/Einspeisung(?)
//	-> Diese Stelle/der Index muss bei der Suche nach den Werten vom Vergleich ausgeklammert werden. <-
//
// Zur Programmentwicklung wird ein Telegramm ins EEPROM gespeichert und von dort für die Dekodierung auf Knopfdruck ausgelesen. 
//	Vorteil: Der Entwicklungsrechner muss nicht am Zähler stehen :)
// HTerm zeichnet die Telegramme aus dem EEPROM auf und kann sie (ggf. manuell geändert) wieder an die Schaltung zurück liefern.
// 
// Nachdem die Dekodierung auf dem Nano lief, wurde der WiFi Teil des Code auf einem Wemos D1 mini lauffähig gemacht. Der Upload auf den 
//	ESP läuft beim Wemos doppelt so schnell wie beim ESP-01 und man muss ihn nicht zum Programmieren erst auf einen Programmer umstecken.

// WiFi libs
#include <WiFiServer.h>
#include <ESP8266WiFi.h>

// benoetigte Bibliotheken für das OLED Display
#include <U8g2lib.h>

// Defines für Debug ----------------------------------------------------------------------------------
//#define RANDOMVALUES					// zufällige Werte für WEB Oberfläche erzeugen
//#define DEBUGWIRKLEISTUNG 103			// Test der Ausgabeformatierung (Dreiecke hoch/runter)
// ----------------------------------------------------------------------------------------------------

// allgemeine Defines
#define MESSAGELENGTH 500				// Anzahl Bytes in smlMessage
#define TELEGRAMMPAUSE 500				// zur Erkennung der Pause zwischen zwei Telegrammen [ms]
#define EESTARTADRESSE 0				// Ablage des Telegramms im EEPROM ab EESTARTADRESSE

// Credentials für das eigene WLAN
#define OwnSSID "OwnSSID"				// -->> hier die eigenen Credentials eintragen <<--
#define OwnPassword "OwnPassword"		// -->> hier die eigenen Credentials eintragen <<--

// Pin Definitionen
#if defined (ARDUINO_ESP8266_WEMOS_D1MINI)	// Wemos D1 R2
// I2C Pins am Wemos D1 R2
#define I2CSDA D1					// beim ESP01 1, beim Wemos 5
#define I2CSCL D2					// beim ESP01 2, beim Wemos 4
#define IRSENSOR RX					// beim ESP01 3, beim Wemos 3
#define TASTER D3					// beim ESP01 0, beim Wemos 0
#define SIGNALLED D4				// beim ESP01 4, beim Wemos 2
#elif defined (ARDUINO_ESP8266_GENERIC)		// ESP-01
// I2C Pins am ESP-01
#define I2CSDA 1					// beim ESP01 1, beim Wemos 5
#define I2CSCL 2					// beim ESP01 2, beim Wemos 4
#define IRSENSOR RX					// beim ESP01 3, beim Wemos 3
#define TASTER 0					// beim ESP01 0, beim Wemos 0
#endif

// OLED instantiieren
#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
#define SCREEN_ADDRESS 0x3C
//									Pins: reset,		 clock,  data
U8X8_SSD1306_128X64_NONAME_HW_I2C display(U8X8_PIN_NONE, I2CSCL, I2CSDA);
// void u8x8_DrawTile(u8x8_t *u8x8, uint8_t x, uint8_t y, uint8_t cnt, uint8_t *tile_ptr); 
// Zeichengenerator "Zeichengenerator_OLED_SSD1306.xlsx"
uint8_t upArrow[] =   { 64, 112, 124, 127, 124, 112, 64, 0 };	// Dreieck nach oben auf dem OLED, Pfeil wäre 0x8C
uint8_t downArrow[] = { 1, 7, 31, 127, 31, 7, 1, 0 };			// Dreieck nach unten auf dem OLED, Pfeil wäre 0x8E

unsigned long DisplayTimer = 0;
bool DisplayContentEin = true;			// Ausgaben auf dem Display de-/aktivieren (PowerSave funktioniert nicht)
#define DISPLAYPERIOD 25000				// OLED Display nach TASTER==LOW für 25 s einschalten
bool clearOnce = true;					// Display einmalig nach Setup löschen

// WiFiServer instantiieren
WiFiServer server(80);
unsigned long clientDelayTimer = 0;
#define CLIENTDELAY 2000				// Werte alle 2 s abfragen und WEB Oberfläche aktualisieren

// globale Variablen
char buffer[20];						// Puffer für formatierte Ausgabe der Zahlen
uint8_t smlMessage[MESSAGELENGTH];		// byte array to store the recorded message for parsing
const uint8_t startSequence[] = { 0x1B, 0x1B, 0x1B, 0x1B, 0x01, 0x01, 0x01, 0x01 }; //start sequence of SML protocol
const uint8_t stopSequence[] =  { 0x1B, 0x1B, 0x1B, 0x1B, 0x1A };					//end sequence of SML protocol
//														Achtung - Offset 11 kann 82, A2, 52 sein, unklar, wovon das abhängt. Einspeisung/Bezug?
//														0	  1     2     3     4     5     6     7     8     9    10   \11/   12    13    14    15    16    17		// Offset im Array
const uint8_t LeadInWirkleistung[] =			   { 0x07, 0x01, 0x00, 0x10, 0x07, 0x00, 0xFF, 0x01, 0x01, 0x62, 0x1B, 0x52, 0x00, 0x55 };							// lead in für "Wirkleistung" value (4 Bytes)
const uint8_t LeadInBezugGesamt[] =			/*180*/{ 0x07, 0x01, 0x00, 0x01, 0x08, 0x00, 0xFF, 0x65, 0x00, 0x01, 0x01, 0x82, 0x01, 0x62, 0x1E, 0x52, 0xFF, 0x59 }; // lead in für "Gesamtverbrauch" value (8 Bytes)
const uint8_t LeadInEinspeisungGesamt[] =	/*280*/{ 0x07, 0x01, 0x00, 0x02, 0x08, 0x00, 0xFF, 0x65, 0x00, 0x01, 0x01, 0x82, 0x01, 0x62, 0x1E, 0x52, 0xFF, 0x59 }; // lead in für "Gesamteinspeisung" value (8 Bytes)

uint16_t startIndex;					// start index for start sequence search
uint16_t stopIndex;						// start index for stop sequence search

// Durch die Verwendung von union erübrigt sich die Shifterei beim Zusammenbasteln der 64-bit Werte aus dem Telegramm.
// nota bene: Auf dem Arduino funktioniert Shift auf 64-bit Variablen ohnehin nicht
union UnionLeistung {
	int32_t  MomentanLeistung;			// variable to hold "Wirkleistung" value
	uint8_t  byteWirkLeistung[4];		// array that holds the extracted 4 byte "Wirkleistung" value
} AktLeistung;
int32_t WirkleistungkW;					// muss auch negativ werden können (Zweirichtungszähler, Bezug positiv/Einspeisung negativ)

union UnionBezug {
	uint64_t GesamtBezug;				// variable to hold "Gesamtbezug" value
	uint8_t  byteGesBezug[8];			// array that holds the extracted 8 byte "Gesamtbezug" value
} Bezug;
double GesamtBezugkWh;					// variable to calulate "Gesamtbezug" in kWh

union UnionEinspeisung {
	uint64_t GesamtEinspeisung;			// variable to hold "Gesamteinspeisung" value
	uint8_t  byteGesEinspeisung[8];		// array that holds the extracted 8 byte "Gesamteinspeisung" value	  
} Einspeisung;
double GesamtEinspeisungkWh;			// variable to calulate "Gesamteinspeisung" in kWh

static uint16_t writePointer = 0;		// inChar an diesem Offset in inputBuffer ablegen - STATIC!!!
bool serialReceived = false;

unsigned long receiveTimer = 0;			// zwischen den Telegrammen ist eine längere Pause (>1 s)

unsigned long activityTimer = 0;		// langsam drehender Propeller auf dem OLED Display
uint8_t propellerState = 0;
char propellerString[] = { '|', '/', '-', '\\' };

// ---------------------------------------------------------------------------------------------------------------------
void setup() {

	//pinMode(IRSENSOR, INPUT_PULLUP);				// Eingang der seriellen SS (RxD)
	pinMode(TASTER, INPUT_PULLUP);					// Taster für Display temporär einschalten

#if defined (ARDUINO_ESP8266_WEMOS_D1MINI)			// LED nur beim Wemos vorhanden
	pinMode(SIGNALLED, OUTPUT);						// signal incoming data on serial
	digitalWrite(SIGNALLED, LOW);
#endif

	// bring up serial port
	Serial.begin(9600);								// Default ist Baudrate + 8N1, das passt für SML (und den MT175)

	// init OLED display, I2C-Adr. 0x3c
	bool result = display.begin();
	display.setFont(u8x8_font_amstrad_cpc_extended_r);
	display.clearDisplay();
	display.drawString(0, 0, "ISKRA MT175");		// show text

	// bring up WiFi
	WiFi.begin(OwnSSID, OwnPassword);				// -->> hier die eigenen Credentials eintragen <<--

	while (WiFi.status() != WL_CONNECTED) {
		delay(500);
		yield();									// für ESP
		Serial.print(".");
	}
	Serial.println("");
	Serial.println("WiFi connected");
	display.drawString(0, 2, "WiFi connected");		// show text

	// Start the server
	server.begin();
	Serial.println("Server started");
	display.drawString(0, 3, "Server started");		// show text

	// Print the IP address
	Serial.print("Use this URL to connect: ");
	Serial.print("http://");
	Serial.print(WiFi.localIP());
	Serial.println("/");
	display.drawString(0, 4, "IP");					// show text
	display.setCursor(3, 4);
	display.print(WiFi.localIP());
	// \bring up WiFi Ende

	// show content of OLED for DISPLAYPERIOD time
	clientDelayTimer = millis();					// direkt nach PowerOn Display erst mal aktiviert lassen...
}


// ---------------------------------------------------------------------------------------------------------------------
void loop()
{
	// WiFi part of loop ----------------------------------------------------------------------------------------------
	if ((millis() - clientDelayTimer) > CLIENTDELAY)	// update values online every CLIENTDELAY milliseconds
	{
#ifdef RANDOMVALUES
		// zufällige Werte für Test
		WirkleistungkW = random(280, 8210);
		GesamtBezugkWh = random(8000, 8200);
		GesamtEinspeisungkWh = random(49200, 51300);
#endif
		WiFiClient client = server.available();
		client.println("HTTP/1.1 200 OK");
		client.println("Content-Type: text/html");
		client.println("Connection: close");			// the connection will be closed after completion of the response
		//client.println("Refresh: 5");					// refresh the page automatically every 5 sec
		client.println();
		client.println("<!DOCTYPE html>");
		//client.println("<html xmlns='http://www.w3.org/1999/xhtml'>");
		client.println("<html>");
		client.println("<head>\n<meta charset='UTF-8'>");
		client.println("<meta http-equiv = \"refresh\" content = \"5\"/>");	// refresh the page automatically every 5 sec
		client.println("<title>Iskra MT175</title>");
		client.println("</head>\n<body>");
#if defined (ARDUINO_ESP8266_GENERIC)
		client.println("<H2>Iskra MT175 ESP-01</H2>");
#else 
		client.println("<H2>Iskra MT175 Wemos</H2>");
#endif
		//client.println("<H3>Humidity / Temperature</H3>");
		client.println("<pre>");
		client.print("Wirkleistung: ");
		dtostrf(WirkleistungkW, 12, 0, buffer);			// Zahl formatieren (12 Stellen, keine Nachkomma, rechtsbündig)
		client.print(buffer);
		client.print(" W ");
		//												// Einspeisung - Dreieck nach oben, Bezug - Dreieck nach unten
		client.println(WirkleistungkW > 0 ? "&#9660;" : WirkleistungkW < 0 ? "&#9650;" : "");
		client.print("Bezug       : ");
		dtostrf(GesamtBezugkWh, 12, 2, buffer);			// Zahl formatieren (12 Stellen inkl. 2 Nachkomma, rechtsbündig)
		client.print(buffer);
		client.println(" kWh");
		client.print("Einspeisung : ");
		dtostrf(GesamtEinspeisungkWh, 12, 2, buffer);	// Zahl formatieren (12 Stellen inkl. 2 Nachkomma, rechtsbündig)
		client.print(buffer);
		client.println(" kWh");
		client.println("</pre>");
		client.print("</body>\n</html>");
		//delay(1000);					// die Pause über delay() verhindert die Aktualisierung des Display (Propeller)...
		clientDelayTimer = millis();	// ...daher Realisierung über "Pause ohne Delay"
	}
	// \WiFi part of loop ---------------------------------------------------------------------------------------------

	// Beim ESP wird das Arduino serialEvent() _nicht_ aufgerufen, das muss explizit im Code passieren
#if defined (ARDUINO_ESP8266_WEMOS_D1MINI) || defined (ARDUINO_ESP8266_GENERIC)
	// Serial part of loop
	if (Serial.available() > 0) serialEvent();		// wenn Daten anliegen, explizit die (Arduino-) ISR aufrufen
	// \Serial part of loop
#endif

	// Taster abfragen. Wenn gedrückt, Display für DISPLAYPERIOD einschalten, dann wieder AUS
	if (digitalRead(TASTER) == LOW)
	{
		DisplayTimer = millis();					// neue Zeit merken

		display.drawString(0, 0, "ISKRA MT175");	// show text

		DisplayContentEin = true;
	}
	else
	{
		// Display nach Ablauf von DISPLAYPERIOD abschalten um Einbrennen zu minimieren
		// Würde man normalerweise mit setPowerSave(0) erledigen, das funktioniert bei meinen Displays aber nicht
		if ((millis() - DisplayTimer) > DISPLAYPERIOD)
		{
			display.clearDisplay();
			DisplayContentEin = false;
		}
	}

	// wenn das Telegramm vollständig erfasst ist kommt eine mehr oder weniger lange Pause vor  dem Nächsten
	if ((serialReceived == true) && ((millis() - receiveTimer) > TELEGRAMMPAUSE))
	{
		findStartSequence();				// Startsequenz suchen
		findStopSequence();					// Ist das Telegramm vollständig?
		serialReceived = false;				// für nächstes Telegramm scharf machen

		// Auswertung
		findConsumptionSequence();			// look for consumption sequence and exctract	- Bezug
		findDeliveredSequence();			// look for delivered sequence and exctract		- Einspeisung
		findPowerSequence();				// look for power sequence and extract			- Momentanleistung
		// Anzeige auf OLED und WEB aktualisieren
		publishMessage();					// do something with the result
	}

#if defined DEBUGWIRKLEISTUNG				// Ausgabe von Debugwerten	
	static unsigned long WkwTimer = 0;
	if (millis() - WkwTimer > CLIENTDELAY)	WirkleistungkW = DEBUGWIRKLEISTUNG;
	if (millis() - WkwTimer > 2 * CLIENTDELAY)	WirkleistungkW = -(DEBUGWIRKLEISTUNG);
	if (millis() - WkwTimer > 3 * CLIENTDELAY) { WirkleistungkW = 0; WkwTimer = millis(); }
	publishMessage();					// do something with the result (Debugausgabe)
#endif

	// Aktivitätsanzeige - rotierender Propeller auf dem OLED Display
	if ((millis() - activityTimer) > 300)
	{
		if (DisplayContentEin == true)
		{
			propellerState++;
			propellerState = propellerState % 4;

			display.setCursor(15, 0);							// set text cursor position
			display.print(propellerString[propellerState]);		// show text
		}
		activityTimer = millis();
	}
}
// ---------------------------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------------------
void findConsumptionSequence() {			// Bezug gesamt - 1.8.0 - 180
	startIndex = 0;

	for (int x = 0; x < sizeof(smlMessage); x++)
	{
		if ((smlMessage[x] == LeadInBezugGesamt[startIndex]) && (startIndex != 11))	// Index 11 ausmaskieren
		{
			startIndex++;
			if (startIndex == sizeof(LeadInBezugGesamt))
			{
				for (int y = 0; y < 8; y++)
				{
					// hier muss für die folgenden 8 Byte hoch gezählt werden
					Bezug.byteGesBezug[7 - y] = smlMessage[x + y + 1];
				}
				startIndex = 0;
			}
		}
		else if (startIndex == 11) startIndex++;	// Wert auf Index 11 ignorieren/nicht vergleichen, der kann 52, 82 oder A2 sein
		else
		{
			startIndex = 0;
		}
	}
	GesamtBezugkWh = Bezug.GesamtBezug / 10000.0; // 10000 impulses per kWh
}

// --------------------------------------------------------------------------------
void findDeliveredSequence() {			// Einspeisung gesamt - 2.8.0 - 280
	startIndex = 0;

	for (int x = 0; x < sizeof(smlMessage); x++)
	{
		if ((smlMessage[x] == LeadInEinspeisungGesamt[startIndex]) && (startIndex != 11))	// Index 11 ausmaskieren
		{
			startIndex++;
			if (startIndex == sizeof(LeadInEinspeisungGesamt))
			{
				for (int y = 0; y < 8; y++)
				{
					// hier muss für die folgenden 8 Byte hoch gezählt werden
					Einspeisung.byteGesEinspeisung[7 - y] = smlMessage[x + y + 1];
				}
				startIndex = 0;
			}
		}
		else if (startIndex == 11) startIndex++;	// Wert auf Index 11 ignorieren/nicht vergleichen, der kann 52, 82 oder A2 sein
		else
		{
			startIndex = 0;
		}
	}
	GesamtEinspeisungkWh = Einspeisung.GesamtEinspeisung / 10000.0; // 10000 impulses per kWh
}

// --------------------------------------------------------------------------------
void findPowerSequence() {					// Wirkleistung momentan
	startIndex = 0;																		// start at position 0 of exctracted SML message

	for (int x = 0; x < sizeof(smlMessage); x++)										// as long as there are elements in the exctracted SML message 
	{
		if ((smlMessage[x] == LeadInWirkleistung[startIndex]) && (startIndex != 11))	// compare with power sequence, Index 11 ausmaskieren
		{
			startIndex++;
			if (startIndex == sizeof(LeadInWirkleistung))								// if complete sequence is found
			{
				// hier muss für die folgenden 4 Byte hoch gezählt werden
				for (int y = 0; y < 4; y++)												// read the next 4 bytes (the actual power value)
				{
					AktLeistung.byteWirkLeistung[3 - y] = smlMessage[x + y + 1];		// store into power array
				}
				startIndex = 0;
			}
		}
		else if (startIndex == 11) startIndex++;	// Wert auf Index 11 ignorieren/nicht vergleichen, der kann 52, 82 oder A2 sein
		else
		{
			startIndex = 0;
		}
	}
	WirkleistungkW = AktLeistung.MomentanLeistung;
}

// --------------------------------------------------------------------------------
void publishMessage()
{
#define Startzeile 2											// Beginn der Ausgaben der Werte

	if (DisplayContentEin == true)
	{
		// Display einmalig nach Setup löschen
		if (clearOnce == true) { display.clearDisplay(); clearOnce = false; }

		display.drawString(15, 0, "ISKRA MT175");				// show text

		display.drawString(0, Startzeile, "Wirkleistung: ");	// show text
		display.setCursor(0, Startzeile + 1);					// set text cursor position	
		dtostrf(WirkleistungkW, 12, 0, buffer);					// Zahl formatieren (12 Stellen, keine Nachkomma, rechtsbündig)
		display.print(buffer);									// show text
		display.print(" W");									// show text

		if (WirkleistungkW < 0) display.drawTile(15, Startzeile + 1, 1, upArrow);			// Einspeisung
		else if (WirkleistungkW > 0) display.drawTile(15, Startzeile + 1, 1, downArrow);	// Bezug
		else  display.drawString(15, Startzeile + 1, " ");

		display.drawString(0, Startzeile + 2, "Bezug: ");		// show text
		display.setCursor(0, Startzeile + 3);
		dtostrf(GesamtBezugkWh, 12, 2, buffer);					// Zahl formatieren (12 Stellen inkl. 2 Nachkomma, rechtsbündig)
		display.print(buffer);									// show text
		display.print(" kWh");									// show text

		display.drawString(0, Startzeile + 4, "Einspeisung: ");	// show text
		display.setCursor(0, Startzeile + 5);					// set text cursor position	
		dtostrf(GesamtEinspeisungkWh, 12, 2, buffer);			// Zahl formatieren (12 Stellen inkl. 2 Nachkomma, rechtsbündig)
		display.print(buffer);									// show text
		display.print(" kWh");									// show text
	}
	// clear the buffers
	memset(smlMessage, 0, sizeof(smlMessage));
}

// --------------------------------------------------------------------------------
void findStartSequence()
{
	startIndex = 0;
	for (int x = 0; x < sizeof(smlMessage); x++)
	{
		//Serial.write(smlMessage[x]);							// debug

		if (smlMessage[x] == startSequence[startIndex])			//in case byte in array matches the start sequence at position 0,1,2...
		{
			startIndex++;
			if (startIndex == sizeof(startSequence))			//all start sequence values have been identified
			{
				startIndex = 0;
			}
		}
		else
		{
			startIndex = 0;
		}
	}
	writePointer = 0;											// start over
}

// --------------------------------------------------------------------------------
void findStopSequence()
{
	stopIndex = 0;

	for (int x = sizeof(startSequence); x < sizeof(smlMessage); x++)	// vergleiche ab nach Startsequenz
	{
		if (smlMessage[x] == stopSequence[stopIndex])
		{
			stopIndex++;
			if (stopIndex == sizeof(stopSequence))
			{
				stopIndex = 0;
			}
		}
		else
		{
			stopIndex = 0;
		}
	}
}

// --------------------------------------------------------------------------------
// Abfrage der seriellen Schnittstelle
 /*
 Arduino:
  SerialEvent occurs whenever new data arives in the hardware serial RX.
  This routine is called each time loop() finishes, so using delay inside loop can
  delay response. Multiple bytes of data may be available.
 ESP8266:
  SerialEvent muss explizit aus der loop() heraus aufgerufen werden, wenn Zeichen
  auf der seriellen SS ankommen.
*/
void serialEvent()
{
#if defined (ARDUINO_ESP8266_WEMOS_D1MINI)
	digitalWrite(SIGNALLED, HIGH);						// show data on Serial
#endif

	serialReceived = true;								// serielle SS hat etwas empfangen
	receiveTimer = millis();

	while (Serial.available())
	{
		uint8_t inChar = (uint8_t)Serial.read();		// Zeichen einlesen

		smlMessage[writePointer] = inChar;				// an inputBuffer anhängen
		writePointer++;

		yield();										// für ESP
	}
#if defined (ARDUINO_ESP8266_WEMOS_D1MINI)
	digitalWrite(SIGNALLED, LOW);						// show data on Serial
#endif
}	// \serialEvent()

